body {
  margin: 0;
  font-family: -apple-system, BlinkMacSystemFont, Segoe UI, Roboto,
    Helvetica Neue, Helvetica, Ubuntu, Arial, sans-serif;
  background: rgba(71, 118, 230, 0.77);
  background: -webkit-linear-gradient(
    0deg,
    rgba(49, 48, 178, 0.62),
    rgba(71, 128, 230, 0.77)
  );
  background: linear-gradient(
    90deg,
    rgba(49, 48, 178, 0.62),
    rgba(71, 128, 230, 0.77)
  );
  display: flex;
  flex-direction: column;
  min-height: 100vh;
}

header {
  font-weight: normal;
  -webkit-font-smoothing: antialiased;
  color: #fff;
  margin-top: 15px;
  margin-left: 20px;
  margin-bottom: 15px;
  white-space: nowrap;
}

.main-grid {
  display: grid;
  flex: 1;
  grid-template-columns: min-content 1fr;
  grid-template-rows: min-content 1fr;
}

@media screen and (max-width: 768px) {
  .sidebar {
    display: none;
  }
  header {
    display: none;
  }
}

.sidebar {
  margin: 0 20px auto;
  position: sticky;
  top: 15px;
  height: fit-content;
  max-height: 100vh;
  overflow-y: auto;
  overscroll-behavior-y: contain;
}

.sidebar ul {
  font-size: 14px;
  list-style-type: none;
  padding: 0px 20px;
  white-space: nowrap;
}

.file-name {
  font-weight: bold;
  font-size: 1.1em;
}

.sidebar ul li {
  margin: 5px 0px;
}

.sidebar button {
  cursor: pointer;
  background: transparent;
  border-color: transparent;
  padding: 0;
}

.sidebar svg.collapsed {
  transform: rotate(-90deg);
}

.sidebar a:link,
a:visited {
  color: inherit;
  text-decoration: none;
}

.sidebar a:hover {
  text-decoration: underline;
}

.title {
  font-size: 2.4em;
  margin: 0;
}

time {
  display: block;
  font-size: 1.2em;
  margin-top: 10px;
}

.wrap {
  padding: 10px;
}

main {
  overflow: auto;
  grid-column: 2;
  grid-row: 1 / span 2;
  padding: 20px;
}

section {
  margin: auto;
  display: inline-block;
  min-width: 50%;
  background: rgba(234, 233, 233, 0.25);
  box-shadow: -1px 9px 15px -5px rgba(77, 77, 77, 0.6);
  border-radius: 15px;
}

.section-header {
  display: flex;
  padding: 0 15px;
  color: rgba(255, 255, 255, 0.9);
  text-shadow: 0px 1px 2px black;
  background: #141e30;
  background: linear-gradient(
    to left,
    rgba(36, 59, 85, 0.67),
    rgba(20, 30, 48, 0.77)
  );
  background: -webkit-linear-gradient(
    right,
    rgba(36, 59, 85, 0.67),
    rgba(20, 30, 48, 0.77)
  );
  border-top-left-radius: 15px;
  border-top-right-radius: 15px;
}

.error-output {
  padding: 0 15px;
}

.code-errors {
}

.style-errors {
  border-top: 3px solid rgba(255, 255, 255, 0.25);
}

.category-heading {
  display: flex; /*Narrow width viewports*/
  -webkit-font-smoothing: antialiased;
  align-items: center;
  margin-top: 7px;
}

.error-instance {
  display: grid;
  background: #fff;
  box-shadow: 0 0 4px rgba(0, 0, 0, 0.1), 0 4px 7px rgba(0, 0, 0, 0.1);
  padding: 8px 16px;
  margin-bottom: 5px;
  border-radius: 3px;
}

.error-link {
  line-height: 35px;
}

.error-link a {
  font-weight: bold;
  text-decoration: none;
  color: inherit;
}
.error-link a:hover {
  text-decoration: underline;
}

.error-link a:visited {
  color: inherit;
}

.code-error-id {
  color: rgba(219, 0, 57, 0.9);
}

.code-error-id a:link {
}

.style-error-id {
  color: rgb(0, 111, 255);
}

.style-error-id a:link {
}

.line-num {
  color: #2d3441;
  font-size: 90%;
}

.grey {
  color: rgb(87, 87, 87);
}

.gbold {
  color: rgb(90, 91, 92);
  font-weight: bold;
}

.highlight-pyta {
  font-weight: bold;
  background: rgba(228, 255, 0, 0.4) !important;
}

.black {
  color: rgb(0, 0, 0);
}

.occurrences,
.shown {
  line-height: 35px;
  padding-left: 20px;
  vertical-align: middle;
}

.more-info {
  color: #000000;
  font-weight: normal;
}

.info {
  font-size: 110%;
  display: flex; /*Narrow width viewports*/
}

.slider {
  margin-left: auto;
}
.slider button {
  background: none;
  border: none;
  cursor: pointer;
}

.slider button:hover {
  filter: brightness(120%);
}

button.collapsed {
  transform: scaleY(-1);
}

.collapsible {
  width: 100%;
  overflow-y: hidden;
}

.hide-and-maintain-width {
  height: 0;
  transition: height 0.15s ease-out;
}

.expanded {
  height: auto;
  transition: height 0.15s ease-out;
}

.content.hide-and-maintain-width,
.message-container.hide-and-maintain-width {
  margin-bottom: 0;
  transition: height 0.3s ease-out;
}

.content.expanded,
.message-container.expanded {
  transition: height 0.3s ease-out;
}

.content {
  margin-bottom: 15px;
}

.message-container.expanded {
  border-top: 1px solid rgba(0, 0, 0, 0.15);
  margin: 8px 0;
}

.message-snippet-container {
  background-color: #f4f8f8;
}

.message-snippet {
  font-size: 1.2em;
  line-height: 1.4em;
  margin: 0;
  padding: 0.25em 0;
}

.message-snippet.hide-and-maintain-width {
  padding: 0;
}

.empty-placeholder {
  font-style: italic;
  padding-left: 20px;
}

/* Footer styling */

footer {
  background: #2d3441;
  color: #ffffff;
  display: flex;
  flex-direction: row;
  padding: 15px;
}

footer a {
  color: #ffffff;
}

footer a:hover {
  color: #86d1ef;
}

footer p {
  margin-top: 0;
}

.footer-text {
  flex: 1;
}

.logos {
  width: 110px;
}

.logo {
  padding: 0 5px;
  text-decoration: none;
}

/*  Pygments css. Edited and reformatted from original output.
    Unfortunately the class names cannot be improved to semantic names.
*/
.pygments-hll {
  background-color: #ffffcc;
} /* ? */

.pygments-c {
  color: #999;
} /* Comment */

.pygments-err {
  /*border: 1px solid #FF0000;*/
} /* Error */

.pygments-k {
  color: #aa22ff;
  font-weight: bold;
} /* Keyword */

.pygments-o {
  color: #666666;
} /* Operator */

.pygments-ch {
  color: #408080;
} /* Comment.Hashbang */

.pygments-cm {
  color: #999;
} /* Comment.Multiline */

.pygments-cp {
  color: #bc7a00;
} /* Comment.Preproc */

.pygments-cpf {
  color: #408080;
} /* Comment.PreprocFile */

.pygments-c1 {
  color: #999;
} /* Comment.Single */

.pygments-cs {
  color: #999;
} /* Comment.Special */

.pygments-gd {
  color: #a00000;
} /* Generic.Deleted */

.pygments-ge {
  font-style: italic;
} /* Generic.Emph */

.pygments-gr {
  color: #ff0000;
} /* Generic.Error */

.pygments-gh {
  color: #000080;
  font-weight: bold;
} /* Generic.Heading */

.pygments-gi {
  color: #00a000;
} /* Generic.Inserted */

.pygments-go {
  color: #888888;
} /* Generic.Output */

.pygments-gp {
  color: #000080;
  font-weight: bold;
} /* Generic.Prompt */

.pygments-gs {
  font-weight: bold;
} /* Generic.Strong */

.pygments-gu {
  color: #800080;
  font-weight: bold;
} /* Generic.Subheading */

.pygments-gt {
  color: #0044dd;
} /* Generic.Traceback */

.pygments-kc {
  color: #008000;
  font-weight: bold;
} /* Keyword.Constant */

.pygments-kd {
  color: #008000;
  font-weight: bold;
} /* Keyword.Declaration */

.pygments-kn {
  color: #008000;
  font-weight: bold;
} /* Keyword.Namespace */

.pygments-kp {
  color: #008000;
} /* Keyword.Pseudo */

.pygments-kr {
  color: #008000;
  font-weight: bold;
} /* Keyword.Reserved */

.pygments-kt {
  color: #b00040;
} /* Keyword.Type */

.pygments-m {
  color: #666666;
} /* Literal.Number */

.pygments-s {
  color: #ba2121;
} /* Literal.String */

.pygments-na {
  color: #7d9029;
} /* Name.Attribute */

.pygments-nb {
  color: #008000;
} /* Name.Builtin */

.pygments-nc {
  color: #0000ff;
  font-weight: bold;
} /* Name.Class */

.pygments-no {
  color: #880000;
} /* Name.Constant */

.pygments-nd {
  color: #aa22ff;
} /* Name.Decorator */

.pygments-ni {
  color: #999999;
  font-weight: bold;
} /* Name.Entity */

.pygments-ne {
  color: #d2413a;
  font-weight: bold;
} /* Name.Exception */

.pygments-nf {
  color: #0000ff;
} /* Name.Function */

.pygments-nl {
  color: #a0a000;
} /* Name.Label */

.pygments-nn {
  color: #0000ff;
  font-weight: bold;
} /* Name.Namespace */

.pygments-nt {
  color: #008000;
  font-weight: bold;
} /* Name.Tag */

.pygments-nv {
  color: #19177c;
} /* Name.Variable */

.pygments-ow {
  color: #aa22ff;
  font-weight: bold;
} /* Operator.Word */

.pygments-w {
  color: #bbbbbb;
} /* Text.Whitespace */

.pygments-mb {
  color: #075cc6;
} /* Literal.Number.Bin */

.pygments-mf {
  color: #075cc6;
} /* Literal.Number.Float */

.pygments-mh {
  color: #075cc6;
} /* Literal.Number.Hex */

.pygments-mi {
  color: #075cc6;
} /* Literal.Number.Integer */

.pygments-mo {
  color: #075cc6;
} /* Literal.Number.Oct */

.pygments-sa {
  color: #059911;
} /* Literal.String.Affix */

.pygments-sb {
  color: #059911;
} /* Literal.String.Backtick */

.pygments-sc {
  color: #059911;
} /* Literal.String.Char */

.pygments-dl {
  color: #059911;
} /* Literal.String.Delimiter */

.pygments-sd {
  color: #059911;
} /* Literal.String.Doc */

.pygments-s2 {
  color: #059911;
} /* Literal.String.Double */

.pygments-se {
  color: #bb6622;
  font-weight: bold;
} /* Literal.String.Escape */

.pygments-sh {
  color: #059911;
} /* Literal.String.Heredoc */

.pygments-si {
  color: #bb6688;
  font-weight: bold;
} /* Literal.String.Interpol */

.pygments-sx {
  color: #008000;
} /* Literal.String.Other */

.pygments-sr {
  color: #bb6688;
} /* Literal.String.Regex */

.pygments-s1 {
  color: #059911;
} /* Literal.String.Single */

.pygments-ss {
  color: #19177c;
} /* Literal.String.Symbol */

.pygments-bp {
  color: #008000;
} /* Name.Builtin.Pseudo */

.pygments-fm {
  color: #0000ff;
} /* Name.Function.Magic */

.pygments-vc {
  color: #19177c;
} /* Name.Variable.Class */

.pygments-vg {
  color: #19177c;
} /* Name.Variable.Global */

.pygments-vi {
  color: #19177c;
} /* Name.Variable.Instance */

.pygments-vm {
  color: #19177c;
} /* Name.Variable.Magic */

.pygments-il {
  color: #666666;
} /* Literal.Number.Integer.Long */
